#include <TInverse.h>

ClassImp(TInverse)

//______________________________________________________________________________
Int_t
TInverse::PseudoInv(
      TArrayD  &a,
const Int_t     m,
const Int_t     n,
const Double_t  frac,
      Int_t    &nrSingular)
{
//  Calculates the pseudo-inverse of matrix A: A^+ = (A^T A)^-1 A^T
//
//  Input:      a          - (m x n)- Matrix A
//              m,n        - dimensions m,n of matrices
//              frac       - if s_max is the largest singular value, s_min is
//                              calculated as s_min = abs(frac)*s_max. All singular
//                              values < s_min are put on zero.
//
//  Output:     a          - (n x m)- Matrix A^+, the pseudo-inverse of A.
//              nrSingular - final number of singular values != 0
//              return     - convergence flag: 0 -> no convergence (in step 2)
//                                             1 -> convergence

  // start with a unit matrix B
  TArrayD b(m*m);
  for (Int_t i = 0; i < m; i++)
    b[i*m+i] = 1.;

  TArrayD d(n);
  TArrayD e(n);

  // step 1: bidiagonalization of A
  Sv1(a,b,d,e,m,n,m);

  // step 2: diagonalization of bidiagonal matrix */
  Int_t fail = Sv2(a,b,d,e,m,n,m);

  if (!fail)
  {
    // step 3: order singular values and perform permutations */
    Sv3(a,b,d,m,n,m);

    // step 4: singular value analysis */
    TArrayD r(m);
    TArrayD x(n*m);
    nrSingular = Sv4(a,b,d,x,r,m,n,m,frac);
    a = x;
  }

  return (fail);
}

//______________________________________________________________________________
Int_t
TInverse::SvDeComp(
      TArrayD &a,
      TArrayD &b,
      TArrayD &x,
      TArrayD &r,
const Int_t    m,
const Int_t    n,
const Double_t frac,
      Int_t   &nrSingular)
{
//  Program SvDeComp performs a single value decomposition of a (m x n) -
//  matrix A, and solves min(A . x - B) . The program expects that m >= n. If
//  this is not the case, a warning is issued.
//
//  Input:  a           - (m x n)- Matrix A
//          b           - m- Vector B, nb -times
//          m,n         - dimensions m,n of matrices
//          frac        - if s_max is the largest singular value, s_min is
//                         calculated as s_min = abs(frac)*s_max. All
//                         singular values < s_min are put on zero.
//
//  Output:  x          - n- Solution Vector, nb -times
//           nrSingular - final number of singular values != 0
//           r          - sum of residuals for vector B, nb -times
//
//  return  - convergence flag:  0 -> convergence    (in step 2)
//                               1 -> no convergence (in step 2)

  if (m < n)
  {
    Error("SvDeComp","matrix A has less rows than columns!");
    return 1;
  }

  const Int_t nb = 1;

  TArrayD d(n);
  TArrayD e(n);

  // step 1: bidiagonaliztion of A
  Sv1(a,b,d,e,m,n,nb);

  // step 2: diagonaliztion of bidiagonal matrix
  Int_t fail = Sv2(a,b,d,e,m,n,nb);

  if (!fail)
  {
    // step 3: order singular values and perform permutations
    Sv3(a,b,d,m,n,nb);

    // step 4: singular value analysis
    r.Set(nb);
    x.Set(nb*n);
    nrSingular = Sv4(a,b,d,x,r,m,n,nb,frac);
  }

  return (fail);
}

//______________________________________________________________________________
Int_t
TInverse::SvDeCompBound(
      TArrayD &a,
      TArrayD &b,
      TArrayD &e,
      TArrayD &d,
      TArrayD &x,
      TArrayD &r,
const Int_t    m,
const Int_t    n,
const Int_t    l,
const Double_t frac,
      Int_t   &nrSingular)
{
//  Program SvDeCompBound performs a single value decomposition of a
//  (m x n) - matrix A, and solves min(A . x - B) under the condition that
//  E . x = D.
//  The program expects that m >= n. If this is not the case, a warning is issued.
//
//  Input:  a           - (m x n)- Matrix A
//          b           - m- Vector B
//          e           - (l x n)- Matrix E
//          d           - l- Vector d
//          m,n,l       - dimensions m,n,l of matrices
//          frac        - if s_max is the largest singular value, s_min is
//                         calculated as s_min = abs(frac)*s_max. All
//                         singular values < s_min are put on zero.
//
//  Output: x           - n- Solution Vector
//          nrSingular  - final number of singular values != 0
//          r           - sum of residuals for vector B
//  return  - convergence flag:  0 -> convergence (in step 2)
//                               1 -> no convergence (in step 2)

  Int_t nminl = n-l;
  if (nminl <= 0)
  {
    Error("SvDeCompBound","too many boundary conditions!");
    return (1);
  }

  TArrayD up(n);
  TArrayD bb(n);

  TArrayD e_loc(e);
  
  // step 1
  for (Int_t i = 0 ; i < l; i++)
  {
    TArrayD v = GetRow(e_loc,n,i);
    DefHouseHolder(v,up[i],bb[i],i,i+1);
    for (Int_t j = i; j < l; j++)
    {
      TArrayD s = GetRow(e_loc,n,j);
      ApplyHouseHolder(v,up[i],bb[i],s,i,i+1);
      if (j == i && n > i)
      {
        for (Int_t k = i+1; k < n; k++)
          s[k] = v[k];
      }
      PutRow(e_loc,s,n,j);
    }
    for (Int_t jj = 0; jj < m; jj++)
    {
      TArrayD s = GetRow(a,n,jj);
      ApplyHouseHolder(v,up[i],bb[i],s,i,i+1);
      PutRow(a,s,n,jj);
    }
  }

  // step 2
  x[0] = d[0]/e_loc[0];
  if (l > 0)
  {
    for (Int_t j = 1; j < l; j++)
    {
      Int_t off_j = j*n;
      x[j] = d[j];
      for (Int_t k = 0; k <= j-1; k++)
        x[j] -= e_loc[off_j+k]*x[k];
      x[j] /= e_loc[off_j+j];
    }
  }

  // step 3
  for (Int_t j = 0; j < m; j++)
  {
    Int_t off_j = j*n;
    for (Int_t k = 0; k < l; k++)
      b[j] -= a[off_j+k]*x[k];
  }

  // step 4
  TArrayD scr = GetSubMat(a,n,m,nminl,Int_t(0),l);
  TArrayD p2(n);
  Int_t fail = SvDeComp(scr,b,p2,r,m,nminl,frac,nrSingular);

  if (!fail)
  {
    nrSingular += l;
    for (Int_t i = 0; i < nminl; i++)
      x[l+i] = p2[i];
    for (Int_t i = l-1; i >= 0; i--)
    {
      TArrayD v = GetRow(e_loc,n,i);
      ApplyHouseHolder(v,up[i],bb[i],x,i,i+1);
    }
  }

  return (fail);
}

//______________________________________________________________________________
void
TInverse::Sv1(
      TArrayD &a,
      TArrayD &b,
      TArrayD &d,
      TArrayD &e,
const Int_t    m,
const Int_t    n,
const Int_t    nb)
{
//  Program Sv1 makes a bidiagonalization of a (m x n) - matrix A,
//  so that A = Q . C . H^T with matrix C bidiagonal.
//
//  Input:  a   - (m x n)- Matrix A
//          b   - nb times m-Vector B
//          m,n - dimensions m,n of matrices
//
//  Output: a   - (n x n)- Matrix h (see above), in the first
//                 n rows of A
//          b   - nb times m- vector which is Q^T . B
//          d   - n- vector containing diagonal elements of matrix C
//          e   - n- vector containing next-to-diagonal elements of matrix c

  TArrayD v(m);
  TArrayD s(m);
  TArrayD ups(n);
  TArrayD bbs(n);

  for (Int_t i = 0; i < n; i++)
  {
    Int_t off_i = i*n;
    // Set up Householder Transformation q(i)

    Double_t up,bb;
    if (i < n-1 || m > n)
    {
      for (Int_t j = 0; j < m; j++)
        v[j] = a[j*n+i];

      DefHouseHolder(v,up,bb,i,i+1);   

      // Apply q(i) to a
      for (Int_t j = i; j < n; j++)
      {
        // householder transformation on j-th column
        Double_t ss = a[off_i+j]*up;
        for (Int_t k = i+1; k < m; k++)
          ss = ss+a[k*n+j]*v[k];
        ss = ss*bb;
        a[off_i+j] = a[off_i+j]+ss*up;
        for (Int_t k = i+1; k < m; k++)
        {
          Int_t off_k = k*n;
          a[off_k+j] = a[off_k+j]+ss*v[k];
        }
      }

      // Apply q(i) to b
      for (Int_t j = 0; j < nb; j++)
      {
        Double_t ss = b[i*nb+j]*up;
        for (Int_t k = i+1; k < m; k++)
          ss += b[k*nb+j]*v[k];
        ss *= bb;
        b[i*nb+j] += ss*up;
        for (Int_t k = i+1; k < m; k++)
          b[k*nb+j] += ss*v[k];
      }
    }
    if (i < n-2)
    {
      // set up Householder Transformation h(i)
      for (Int_t j = 0; j < n; j++)
        v[j] = a[off_i+j];

      DefHouseHolder(v,up,bb,i+1,i+2);

      // save h(i)
      ups[i] = up;
      bbs[i] = bb;

      // apply h(i) to a
      for (Int_t j = i; j < m; j++)
      {
        Int_t off_j = j*n;
        // householder transformation on j-th row
        Double_t ss = a[off_j+i+1]*up;
        for (Int_t k = i+2; k < n; k++)
          ss = ss+a[off_j+k]*v[k];
        ss = ss*bb;
        a[off_j+i+1] = a[off_j+i+1]+ss*up;
        for (Int_t k = i+2; k < n; k++)
          a[off_j+k] = a[off_j+k]+ss*v[k];

        // save elements i+2,...in row j of matrix a
        if (j == i)
        {
          for (Int_t k = i+2; k < n; k++)
            a[off_j+k] = v[k];
        }
      }
    }
  }

  // copy diagonal of transformed matrix a to d and upper parallel a to e
  if (n > 1)
  {
    for (Int_t i = 1; i < n; i++)
    {
      Int_t off_i = i*n;
      d[i] = a[off_i+i];
      e[i] = a[off_i-n+i];
    }
  }
  d[0] = a[0];
  e[0] = 0.;

  // construct product matrix h = h(1)*h(2)*...*h(n-1), h(n-1) = I

  for (Int_t i = n-1; i >= 0; i--)
  {
    Int_t off_i = i*n;
    Int_t off_i1 = (i+1)*n;
    if (i <= n-2)
    {
      for (Int_t j = 0; j < n; j++)
        v[j] = a[off_i+j];
    }

    for (Int_t k = 0; k < n; k++)
      a[off_i+k] = 0.;
    a[off_i+i] = 1.;
    if (i < n-2)
    {
      for (Int_t k = i; k < n; k++)
      {
        // householder transformation on k-th column
        Double_t ss = a[off_i1+k]*ups[i];
        for (Int_t j = i+2; j < n; j++)
          ss = ss+a[j*n+k]*v[j];
        ss = ss*bbs[i];
        a[off_i1+k] += ss*ups[i];
        for (Int_t j = i+2; j < n; j++)
          a[j*n+k] += ss*v[j];
      }
    }
  }

}

#define ZEROO 0.0

//______________________________________________________________________________
Int_t
TInverse::Sv2(
      TArrayD &a,
      TArrayD &b,
      TArrayD &d,
      TArrayD &e,
const Int_t    m,
const Int_t    n,
const Int_t    nb)
{
//  Program Sv2 diagonalizes in an iterativ fashion the bidiagonal
//  matrix c, so that S' = U'^T . C . V'  is diagonal.
//
//  Input:  a   - (n x n)- Matrix h (in the first n rows of a)
//          b   - nb times m- vector which is  Q^T . B
//          d   - n- vector containing diagonal elements of matrix C
//          e   - n- vector containing next-to-diagonal elements of matrix C
//          m,n - dimensions m,n of matrices
//
//  Output:  a  - (n x n)- Matrix H . V' (see above), in the first
//                n rows of A
//           b  - nb times m- vector which is U'^T . Q^T . B
//           d  - n- vector containing diagonal elements of matrix S'
//  return   - convergence flag:  0 -> no convergence
//                                1 -> convergence

  Int_t fail   = 0;
  Int_t niterm = 10*n;
  Int_t niter  = 0;
  Double_t bmx = d[0];

  if (n > 1)
  {
    for (Int_t i = 1; i < n; i++)
      bmx = TMath::Max(TMath::Abs(d[i])+TMath::Abs(e[i]),bmx);
  }

  for (Int_t k = n-1; k >= 0; k--)
  {
    loop:
      if (k != 0)
      {
        // since d[k] == 0 perform Givens transform
        // with result e[k] = 0
        if ((bmx+d[k])-bmx == ZEROO)
          Sv2_1(a,d,e,n,k);

        // find l (1 <= l <=k) so that either e[l] = 0
        // or d[l-1] = 0. In the latter case transform
        // e[l] to zero. In both cases the matrix
        // splits and the bottom right minor begins with row l.
        // If no such l is found set l = 1.

        Int_t elzero = 0;
        Int_t l;
        for (Int_t ll = k; ll >= 0 ; ll--)
        {
          l = ll;
          if (l == 0)
          {
            elzero = 0;
            break;
          }
          else if ((bmx-e[l])-bmx == ZEROO)
          {
            elzero = 1;
            break;
          }
          else if ((bmx+d[l-1])-bmx == ZEROO)
            elzero = 0;
        }
        if (l > 0 && !elzero)
          Sv2_2(b,d,e,nb,k,l);
        if (l != k)
        {
          // one more QR pass with order k
          Sv2_3(a,b,d,e,n,nb,k,l);
          niter++;
          if (niter <= niterm) goto loop;
          fail = 1;
        }
     }

    if (d[k] < 0.)
    {
     // for negative singular values perform change of sign
      d[k] = -d[k];
      for (Int_t j = 0; j < n; j++)
      {
        Int_t off_j = j*n;
        a[off_j+k] = -a[off_j+k];
      }
    }   
    // order is decreased by one in next pass
  }   

  return (fail);
}

//______________________________________________________________________________
void
TInverse::Sv2_1(
      TArrayD &a,
      TArrayD &d,
      TArrayD &e,
const Int_t    n,
const Int_t    k)
{
  for (Int_t i = k-1; i >= 0;i--)
  {
    Double_t h,cs,sn;
    if (i == k-1)
      DefAplGivens(d[i],e[i+1],cs,sn);
    else
      DefAplGivens(d[i],h,cs,sn);
    if (i > 0)
    {
      h = 0.;
      ApplyGivens(e[i],h,cs,sn);
    }
    for (Int_t j = 0; j < n; j++)
    {
      Int_t off_j = j*n;
      ApplyGivens(a[off_j+i],a[off_j+k],cs,sn);
    }
  }
}

//______________________________________________________________________________
void
TInverse::Sv2_2(
      TArrayD &b,
      TArrayD &d,
      TArrayD &e,
const Int_t    nb,
const Int_t    k,
const Int_t    l)
{
  for (Int_t i = l; i <= k; i++)
  {
    Double_t h,cs,sn;
    if (i == l)
      DefAplGivens(d[i],e[i],cs,sn);
    else
      DefAplGivens(d[i],h,cs,sn);
    if (i < k)
    {
      h = 0.;
      ApplyGivens(e[i+1],h,cs,sn);
    }
    for (Int_t j = 0; j < nb; j++)
      ApplyGivens(cs,sn,b[i*nb+j],b[(l-1)*nb+j]);
  }
}

//______________________________________________________________________________
void
TInverse::Sv2_3(
      TArrayD &a,
      TArrayD &b,
      TArrayD &d,
      TArrayD &e,
const Int_t    n,
const Int_t    nb,
const Int_t    k,
const Int_t    l)
{
  // determine shift parameter
  Double_t f = ((d[k-1]-d[k])*(d[k-1]+d[k])+(e[k-1]-e[k])*(e[k-1]+e[k]))/
                (2.*e[k]*d[k-1]);
  Double_t g;
  if (TMath::Abs(f) > 1.e+10)
    g = TMath::Abs(f);
  else
    g = TMath::Sqrt(1.+f*f);

  Double_t t;
  if (f >= 0.)
    t = f+g;
  else
    t = f-g;

  f = ((d[l]-d[k])*(d[l]+d[k])+e[k]*(d[k-1]/t-e[k]))/d[l];

  Double_t h,cs,sn;
  for (Int_t i = l; i <= k-1; i++)
  {
    Int_t off_i = i*n;
    if (i == l)
      // define r[l]
      DefGivens(f,e[i+1],cs,sn);
    else
      // define r[i]
      DefAplGivens(e[i],h,cs,sn);

    ApplyGivens(d[i],e[i+1],cs,sn);
    h = 0.;
    ApplyGivens(h,d[i+1],cs,sn);
    for (Int_t j = 0; j < n; j++)
    {
      Int_t off_j = j*n;
      ApplyGivens(a[off_j+i],a[off_j+i+1],cs,sn);
    }
    // define t[i]
    DefAplGivens(d[i],h,cs,sn);
    ApplyGivens(e[i+1],d[i+1],cs,sn);
    if (i < k-1)
    {
      h = 0.;
      ApplyGivens(h,e[i+2],cs,sn);
    }
    for (Int_t j = 0; j < nb; j++)
    {
      Int_t off_i2 = i*nb;
      ApplyGivens(b[off_i2+j],b[off_i2+nb+j],cs,sn);
    }
  }
}

//______________________________________________________________________________
void
TInverse::Sv3(
      TArrayD &a,
      TArrayD &b,
      TArrayD &d,
const Int_t    m,
const Int_t    n,
const Int_t    nb)
{
//  Program Ssv3 performs a permutation transformation on the diagonal
//  matrix S, so that matrix S'' = U''^T . S . V''  has diagonal elements
//  ordered such that they do not increase.
//
//  Input:  a   - (n x n)- Matrix H . V' (in the first n rows of a)
//          b   - nb times m- vector which is U'^T . Q^T . B
//          d   - n- vector containing diagonal elements of matrix S'
//          m,n - dimensions m,n of matrices
//
// Output:  a   - (m x n)- Matrix H . V' . V'' (see above), in the first
//                 n rows of A
//          b   - m- vector which is U''^T . U'^T . Q^T . B
//          d   - n- vector containing diagonal elements of matrix S''

  // order singular values

  if (n > 1)
  {
    while (1)
    {
      Int_t found = 0;
      {
        Int_t i = 1;
        while (!found && i < n)
        {
          if (d[i] > d[i-1]) found = 1;
          i++;
        } 
      }
      if (!found) break;
      for (Int_t i = 1; i < n; i++)
      {
        Double_t t = d[i-1];
        Int_t k = i-1;
        for (Int_t j = i; j < n; j++)
        {
          if (t < d[j])
          {
            t = d[j];
            k = j;
          }
        }
        if (k != i-1)
        {
          // perform permutation on singular values
          d[k]   = d[i-1];
          d[i-1] = t;
          // perform permutation on matrix a
          for (Int_t j = 0; j < n; j++)
          {
            Int_t off_j = j*n;
            t            = a[off_j+k];
            a[off_j+k]   = a[off_j+i-1];
            a[off_j+i-1] = t;
          }
          // perform permutation on vector b
          for (Int_t j = 0; j < nb; j++)
          {
            Int_t off_k  = k*nb;
            Int_t off_i1 = (i-1)*nb;
            t = b[off_k+j];
            b[off_k+j]  = b[off_i1+j];
            b[off_i1+j] = t;
          }
        }
      }
    }
  }
}

#define EPSILON 1.0e-30

//______________________________________________________________________________
Int_t
TInverse::Sv4(
      TArrayD &a,
      TArrayD &b,
      TArrayD &d,
      TArrayD &x,
      TArrayD &r,
const Int_t    m,
const Int_t    n,
const Int_t    nb,
const Double_t frac)
{
//  Program Sv4 performs a permutation transformation on the diagonal
//  matrix S, so that matrix S'' = U''^T . S . V''  has diagonal elements
//  ordered such that they do not increase.
//
//  Input:  a    - (n x n)- Matrix H . V' . V'' (in the first n rows of a)
//          b    - nb times m- vector which is  matrix U''^T . U'^T . Q^T . B
//          d    - n- vector containing diagonal elements of matrix S'',so
//                  a vector containing singular values of matrix a,
//                  in a non-ascending order
//          m,n  - dimensions m,n of matrices
//          frac - if s_max is the largest singular value, s_min is
//                  calculated as s_min = abs(frac)*s_max. All singular
//                  values < s_min are put on zero.
//                  if abs(frac) < 10e-30 than the program defaults to frac = 10e-30
//
//  Output: x    - nb times n- Solution Vector
//          r    - nb- vector of sum of the residuals for nb times vector B
//  return       - final number of singular values

  Double_t fract = TMath::Abs(frac);   
  if (fract < EPSILON) fract = EPSILON;

  Double_t sinmax = 0.;
  for (Int_t i = 0; i < n; i++)
    sinmax = TMath::Max(sinmax,d[i]);
  Double_t sinmin = sinmax*fract;

  Int_t kk = n-1;
  for (Int_t i = 0; i < n; i++)
  {
    if (d[i] <= sinmin)
    {
      kk = i-1;
      break;
    }
  }

  for (Int_t i = 0; i < m; i++)
  {
    Int_t off_i = i*nb;
    if (i <= kk)
      for (Int_t j = 0; j < nb; j++)
        b[off_i+j] /= d[i];
    else
    {
      for (Int_t j = 0; j < nb; j++)
      {
        if (i == kk+1)
          r[j] = b[off_i+j]*b[off_i+j];
        else
          r[j] += b[off_i+j]*b[off_i+j];
        if (i < n) b[off_i+j] = 0.;
      }
    }   
  }

  for (Int_t i = 0; i < n; i++)
  {
    Int_t off_i  = i*n;
    Int_t off_ib = i*nb;
    for (Int_t j = 0; j < nb; j++)
    {
      x[off_ib+j] = 0.;
      for (Int_t k = 0; k < n; k++)
        x[off_ib+j] += a[off_i+k]*b[k*nb+j];
    }
  }

  return (kk+1);
}

//______________________________________________________________________________
TArrayD
TInverse::GetColumn(
      TArrayD &a,
const Int_t    m,
const Int_t    n,
const Int_t    i)
{
//  Program GetColumn takes the ith-column vector c out of matrix a.
//
//  Input:  a  - (m x n)- Matrix A
//          m  - number of rows in matrix
//          i  - column index i of matrix A
//
//  Output: c  - m- vector C

  TArrayD c(m);
  for (Int_t j = 0; j < m; j++)
    c[j] = a[j*n+i];
  return c;
}

//______________________________________________________________________________
void
TInverse::PutColumn(
      TArrayD &a,
      TArrayD &c,
const Int_t    m,
const Int_t    n,
const Int_t    i)
{
//  Program PutColumn replaces column i of matrix with vector C
//
//  Input:  a   - (m x n)- Matrix A
//          c   - m- vector C
//          m   - number of rows in matrix
//          i   - column index i of matrix A
//
//  Output: a   - (m x n)- Matrix A with column i replaced by C

  for (Int_t j = 0; j < m; j++)
    a[j*n+i] = c[j];
}

//______________________________________________________________________________
TArrayD
TInverse::GetRow(
      TArrayD &a,
const Int_t    n,
const Int_t    i)
{
//  Program GetRow takes the ith-column vector c out of matrix a.
//
//  Input:   a  - (m x n)- Matrix A
//           n  - number of columns in matrix
//           i  - row index i of matrix A
//
//  Output:  r  - n- vector R

  TArrayD r(n);
  Int_t off_i = i*n;
  for (Int_t j = 0; j < n; j++)
    r[j] = a[off_i+j];
  return r;
}

//______________________________________________________________________________
void
TInverse::PutRow(
      TArrayD &a,
      TArrayD &r,
const Int_t    n,
const Int_t    i)
{
//  Program PutRow replaces row i of matrix i with vector C
//
//  Input:  a   - (m x n)- Matrix A
//          r   - n- vector C
//          n   - number of columns in matrix
//          i   - row index i of matrix A
//
//  Output: a   - (m x n)- Matrix A with row i replaced by C

  Int_t off_i = i*n;
  for (Int_t j = 0; j < n; j++)
    a[off_i+j] = r[j];
}

//______________________________________________________________________________
void
TInverse::DefGivens(
const Double_t v1,
const Double_t v2,
      Double_t &c,
      Double_t &s)
{
//  Program DefGivens defines a Givens-rotation by calculating 2
//  constants c and s.
//
//  Input:  v1
//          v2 - vector components that define the Givens- rotation
//
//  Output: c,
//          s  - constants used in the rotation

  const Double_t a1 = TMath::Abs(v1);
  const Double_t a2 = TMath::Abs(v2);
  if (a1 > a2)
  {
    Double_t w = v2/v1;
    Double_t q = TMath::Sqrt(1.+w*w);
    c = 1./q;
    if (v1 < 0.) c = -c;   
    s = c*w;
  }
  else
  {
    if (v2 != 0)
    {
      Double_t w = v1/v2;
      Double_t q = TMath::Sqrt(1.+w*w);
      s = 1./q;
      if (v2 < 0.) s = -s;   
      c = s*w;
    }
    else
    {
      c = 1.;
      s = 0.;
    }
  }
}

//______________________________________________________________________________
void
TInverse::DefAplGivens(
Double_t &v1,
Double_t &v2,
Double_t &c,
Double_t &s)
{
//  Program DefAplGivens defines a Givens-rotation by calculating 2
//  constants c and s.
//
//  Input:  v1
//          v2   - vector components that define the Givens- rotation
//
//  Output: c,
//          s   - constants used in the rotation

  const Double_t a1 = TMath::Abs(v1);
  const Double_t a2 = TMath::Abs(v2);
  if (a1 > a2)
  {
    const Double_t w = v2/v1;
    const Double_t q = TMath::Sqrt(1.+w*w);
    c = 1./q;
    if (v1 < 0.) c = -c;   
    s  = c*w;
    v1 = a1*q;
    v2 = 0.;
  }
  else
  {
    if (v2 != 0)
    {
      const Double_t w = v1/v2;
      const Double_t q = TMath::Sqrt(1.+w*w);
      s = 1./q;
      if (v2 < 0.) s = -s;   
      c  = s*w;
      v1 = a2*q;
      v2 = 0.;
    }
    else
    {
      c = 1.;
      s = 0.;
    }
  }
}

//______________________________________________________________________________
void
TInverse::ApplyGivens(
      Double_t &z1,
      Double_t &z2,
const Double_t  c,
const Double_t  s)
{
//  Program ApplyGivens applies a Givens transformation as defined by (c,s)
//  to a vector.
//
//  Input:  z1,
//          z2,  - vector components that have to be transformed
//          c,
//          s    - constants of a Givens rotation that have been made by
//                a call to DefGivens or DefAplGivens
//
//  Output: z1,
//          z2   - vector components of the transformed vector

  const Double_t w  = z1*c+z2*s;
  z2 = -z1*s+z2*c;
  z1 = w;
}

//______________________________________________________________________________
void
TInverse::DefHouseHolder(
      TArrayD  &v,
      Double_t &up,
      Double_t &b,
const Int_t    lp,
const Int_t    l)
{
//  Program DefHouseHolder defines a Householder-transformation.
//
//  Input:  v,     - n- vector v
//          n,lp,l - vector indices
//
//  Output: up,    - vector element u(p)
//          b      - coefficient b

  Int_t    n = v.GetSize();
  Double_t c= TMath::Abs(v[lp]);
  for (Int_t i = l; i < n; i++)
    c = TMath::Max(TMath::Abs(v[i]),c);

  up = 0.0;
  b  = 0.0;
  if (c <= 0.) return;

  const Double_t c1 = 1./c;
  Double_t sd = v[lp]*c1*v[lp]*c1;
  for (Int_t i = l; i < n; i++)
    sd += v[i]*c1*v[i]*c1;

  Double_t vpprim = c*TMath::Sqrt(sd);
  if (v[lp] > 0.) vpprim = -vpprim;
  up = v[lp]-vpprim;
  b = 1./(vpprim*up);
}

//______________________________________________________________________________
void
TInverse::ApplyHouseHolder(
      TArrayD  &v,
const Double_t  up,
const Double_t  b,
      TArrayD  &c,
const Int_t     lp,
const Int_t     l)
{
//  Program DefHouseHolder defines a Householder-transformation.
//
//  Input:  v,     - n- vector v
//          up,b   - Householder parameters as determined by DefHouseHolder
//          lp,l   - vector indices
//          c      - n- vector to be transformed
//
//  Output: c      - n- vector transformed

  Int_t n = v.GetSize();
  Double_t s = c[lp]*up;
  for (Int_t i = l; i < n; i++)
    s += c[i]*v[i];

  s = s*b;
  c[lp] += s*up;
  for (Int_t i = l; i < n; i++)
    c[i] += s*v[i];
}

//______________________________________________________________________________
TArrayD
TInverse::GetSubMat(
      TArrayD &a,
const Int_t n,
const Int_t k,
const Int_t l,
const Int_t m0,
const Int_t n0)
{
//  Program GetSubMat gets sub-matrix S (k x l) out of matrix A (m x n).
//
//  Input:  a     - matrix A
//          k,l   - dimensions k,l of matrix S
//          m0,n0 - indices m0, n0 in A, so that A[m0][n0] = S[0][0]
//
//  Output: s     - (k x l)- matrix S

  TArrayD s(k*l);
  for (Int_t i = 0; i < k; i++)
  {
    Int_t off1 = i*l;
    Int_t off2 = (m0+i)*n+n0;
    for (Int_t j = 0; j < l; j++)
      s[off1+j] = a[off2+j];
  }

  return s;
}

//______________________________________________________________________________
void
TInverse::PutSubMat(
      TArrayD &a,
      TArrayD &s,
const Int_t    n,
const Int_t    k,
const Int_t    l,
const Int_t    m0,
const Int_t    n0,
const Double_t fact)
{
//  Program PutSubMat puts sub-matrix S (k x l) into matrix A (m x n).
//
//  Input:  a     - matrix A
//          s     - (k x l)- matrix S
//          k,l   - dimensions k,l of matrix S
//          m0,n0 - indices m0, n0 in A, so that A[m0][n0] = S[0][0]
//          fact  - factor by which S is scaled before inserting
//
//  Output: a     - (m x n)- matrix A

  for (Int_t i = 0; i < k; i++)
  {
    Int_t off1 = i*l;
    Int_t off2 = (m0+i)*n+n0;
    for (Int_t j = 0; j < l; j++)
      a[off2+j] = fact*s[off1+j];
  }
}

//______________________________________________________________________________
Int_t TInverse::Inverse(
TArrayD &a)
{
//  Program Inverse inverts (n x n) - matrix A, using a LU decomposition
//
//  Input:	a	- (n x n)- Matrix A
//  		n	- dimensions n of matrices
//
//  Output:	a	- (n x n)- inverse Matrix A^(-1)
//		return	- 0 inversion succesful
//			- 1 inversion failed

   Int_t    n   = TMath::Sqrt(Long_t(a.GetSize()));
   Double_t *pa = a.GetArray();

   Int_t *index = new Int_t[n];

   Double_t d = 0.0;
   Int_t fail = LUDecompose(pa,n,index,d);
   if (!fail) LUBacksub(pa,n,index);

   delete [] index;

   return (fail);
}

#define	TINY 1.0e-20

//______________________________________________________________________________
Int_t TInverse::LUDecompose(
Double_t *a,
Int_t    n,
Int_t    *index,
Double_t d)
{
//  Program LUDecompose performs on  (n x n) - matrix A a LU decomposition.
//  MAtrix A is overwritten in the process.
//
//  Input:	a	- (n x n)- Matrix A
//  		n	- dimensions n of matrices
//
//  Output:	a	- (n x n)- LU decomposed Matrix A
//		index	- n- vector containing the row permutations effected the
//			  partial pivoting.
//		d	- value equal +1/-1 depending on wether the number
//			  of row interchages was  even or odd, respectively.
//		return	- 0 inversion succesful
//			- 1 inversion failed

  Double_t *vv = new Double_t[n*n];

  Double_t aamax;
  d = 1.;
  for (Int_t i = 0; i < n; i++)
  {
    Int_t off_i = i*n;
    aamax = 0.;
    for (Int_t j = 0; j < n; j++)
      if (TMath::Abs(a[off_i+j]) > aamax) aamax = TMath::Abs(a[off_i+j]);
    if (aamax <= 0.)
    {
      delete [] vv;
      return (1);
    }
    vv[i] = 1./aamax;
  }

  Int_t imax;
  for (Int_t j = 0; j < n; j++)
  {
    Int_t off_j = j*n;
    for (Int_t i = 0; i <= j-1; i++)
    {
      Int_t off_i = i*n;
      Double_t sum = a[off_i+j];
      for (Int_t k = 0; k <= i-1; k++)
        sum = sum-a[off_i+k]*a[k*n+j];
      a[off_i+j] = sum;
    }
    aamax = 0.0;
    for (Int_t i = j; i < n; i++)
    {
      Int_t off_i = i*n;
      Double_t sum = a[off_i+j];
      for (Int_t k = 0; k <= j-1; k++)
        sum = sum-a[off_i+k]*a[k*n+j];
      a[off_i+j] = sum;
      Double_t dum = vv[i]*TMath::Abs(sum);
      if (dum >= aamax)
      {
        imax = i;
        aamax = dum;
      }
    }
    if (j != imax)
    {
      for (Int_t k = 0; k < n; k++)
      {
        Int_t off_imax = imax*n;
        Double_t dum = a[off_imax+k];
        a[off_imax+k] = a[off_j+k];
        a[off_j+k] = dum;
      }
      d = -d;
      vv[imax] = vv[j];
    }
    index[j] = imax;

//    if (a[off_j+j] == 0.) a[off_j+j] = TINY;
    if (a[off_j+j] == 0.)
    {
      delete [] vv;
      return (1);
    }
    if (j != n-1)
    {
      Double_t dum = 1./a[off_j+j];
      for (Int_t i = j+1; i < n; i++)
        a[i*n+j] *= dum;
    }
  }

  delete [] vv;

  return (0);
}
	
//______________________________________________________________________________
void TInverse::LUBacksub(
Double_t *a,
Int_t    n,
Int_t    *index)
{
//  Program LUBacksub calulates of (n x n) - LU matrix A its inverse.
//  Matrix A is overwritten in the process.
//
//  Input:	a	- (n x n)- LU Matrix A
//  		n	- dimensions n of matrices
//		index	- n- vector containing the row permutations effected the
//			  partial pivoting.
//
//  Output:	a	- (n x n)- inverse Matrix A^(-1)

  Double_t *b = new Double_t[n*n];

  for (Int_t i = 0; i < n; i++)
  {
    Int_t off_i = i*n;
    for (Int_t j = 0; j < n; j++)
      b[off_i+j] = 0.0;
    b[off_i+i] = 1.0;
  }
   
  for (Int_t ic = 0; ic < n; ic++)
  {
    Int_t ii = -1;
    for (Int_t i = 0; i < n ; i++)
    {
      Int_t off_i = i*n;
      Int_t ll = index[i];
      Double_t sum = b[ll*n+ic];
      b[ll*n+ic] = b[off_i+ic];
      if (ii != -1)
      {
        for (Int_t j = ii; j <= i-1; j++)
          sum = sum-a[off_i+j]*b[j*n+ic];
      }
      else if (sum != 0)
        ii = i;
      b[off_i+ic] = sum;
    }
    for (Int_t i = n-1; i >= 0; i--)
    {
      Int_t off_i = i*n;
      Double_t sum = b[off_i+ic];
      if (i < n-1)
      {
        for (Int_t j = i+1; j < n; j++)
          sum = sum-a[off_i+j]*b[j*n+ic];
      }
      b[off_i+ic] = sum/a[off_i+i];
    }
  }

  for (Int_t i = 0; i < n; i++)
  {
    Int_t off_i = i*n;
    for (Int_t j = 0; j < n; j++)
      a[off_i+j] = b[off_i+j];
  }

  delete [] b;
}

//______________________________________________________________________________
Int_t TInverse::InversePosDef(
TArrayD &a)
{
//  Program InversePosDef inverts a positive definite (n x n) - matrix A,
//  using the Cholesky decomposition
//
//  Input:	a	- (n x n)- Matrix A
//  		n	- dimensions n of matrices
//
//  Output:	a	- (n x n)- inverse Matrix A^(-1)
//		return	- 0 inversion succesful
//			- 1 inversion failed

   Int_t    n   = TMath::Sqrt(Long_t(a.GetSize()));
   Double_t *pa = a.GetArray();

   Double_t *u = new Double_t[n*n];

   // step 1: Cholesky decomposition
   if (Pdcholesky(pa,u,n))
   {
     delete [] u;
     return (1);
   }

   Int_t off_n = (n-1)*n;
   for (Int_t i = 0; i < n; i++)
   {
     Int_t off_i = i*n;

   // step 2: Forward substitution
     for (Int_t l = i; l < n; l++)
     {
       if (l == i)
         pa[off_n+l] = 1./u[l*n+l];
       else
       {
         pa[off_n+l] = 0.;
         for (Int_t k = i; k <= l-1; k++)
           pa[off_n+l] = pa[off_n+l]-u[k*n+l]*pa[off_n+k];
         pa[off_n+l] = pa[off_n+l]/u[l*n+l];
       }
     }

   // step 3: Back substitution
     for (Int_t l = n-1; l >= i; l--)
     {
       Int_t off_l = l*n;
       if (l == n-1)
         pa[off_i+l] = pa[off_n+l]/u[off_l+l];
       else
       {
         pa[off_i+l] = pa[off_n+l];
         for (Int_t k = n-1; k >= l+1; k--)
           pa[off_i+l] = pa[off_i+l]-u[off_l+k]*pa[off_i+k];
         pa[off_i+l] = pa[off_i+l]/u[off_l+l];
       }
     }
   }

   // Fill lower triangle symmetrically
   if (n > 1)
   {
     for (Int_t i = 0; i < n; i++)
     {
       for (Int_t l = 0; l <= i-1; l++)
         pa[i*n+l] = pa[l*n+i];
     }
   }

   delete [] u;

   return (0);
}

Int_t TInverse::Pdcholesky(
Double_t *a,
Double_t *u,
Int_t    n)
{
//______________________________________________________________________________
//  Program Pdcholesky inverts a positiv definite (n x n) - matrix A,
//  using the Cholesky decomposition
//
//  Input:	a	- (n x n)- Matrix A
//  		n	- dimensions n of matrices
//
//  Output:	u	- (n x n)- Matrix U so that U^T . U = A
//		return	- 0 decomposition succesful
//			- 1 decomposition failed

  for (Int_t i = 0; i < n; i++)
  {
    Int_t off_i = i*n;
    for (Int_t j = 0; j < n; j++)
      u[off_i+j] = 0.;
  }

  for (Int_t k = 0; k < n; k++)
  {
    Double_t s = 0.;
    Int_t off_k = k*n;
    for (Int_t j = k; j < n; j++)
    {
      if (k > 0)
      {
        s = 0.;
        for (Int_t l = 0; l <= k-1; l++)
        {
          Int_t off_l = l*n;
          s += u[off_l+k]*u[off_l+j];
        }
      }
      u[off_k+j] = a[off_k+j]-s;
      if (k == j)
      {
        if (u[off_k+j] <= 0)
        {
          Error("Pdcholesky","matrix not pos definite: %f\n",u[off_k+j]);
          return (1);
        }
        u[off_k+j] = TMath::Sqrt(u[off_k+j]);
      }
      else
        u[off_k+j] = u[off_k+j]/u[off_k+k];
    }
  }
  return (0);
}
