//
// This file contains the class InputDialog.
// An InputDialog object prompts for an input string using a simple
// dialog box. The InputDialog class is also a good example of how
// to use the ROOT GUI classes via the interpreter. Since interpreted
// classes can not call virtual functions via base class pointers, all
// GUI objects are used by composition instead of by inheritance.
//
// This file contains also some utility functions that use
// the InputDialog class to either get a string, integer or
// floating point number. There are also two functions showing
// how to use the file open and save dialogs. The utility functions are:
//
//  char *OpenFileDialog()
//  char *SaveFileDialog()
//  char *GetStringDialog( char *prompt,  char *defval)
// Int_t GetIntegerDialog( char *prompt, Int_t defval)
// Float_t GetFloatDialog( char *prompt, Float_t defval)
//
// To use the InputDialog class and the utility functions you just
// have to load the dialogs.C file as follows:
// .L dialogs.C
//
// Now you can use them like:
// {
//     char *file = OpenFileDialog();
//    Int_t run   = GetIntegerDialog("Give run number:", 0);
//    Int_t event = GetIntegerDialog("Give event number:", 0);
//    printf("analyse run %d, event %d from file %s\n", run ,event, file);
// }
//

///////////////////////////////////////////////////////////////////////////
//                                                                       //
// Input Dialog Widget                                                   //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>

#include <TGFrame.h>
#include <TGLabel.h>
#include <TGTextEntry.h>
#include <TGTextBuffer.h>
#include <TGFileDialog.h>
#include <TGButton.h>
#include <TGWindow.h>
#include <TInterpreter.h>
#include <TList.h>
#include <TMath.h>

#include <TROOT.h>
#include <TApplication.h>
#include <TVirtualX.h>

#include "dialogs.h"

ClassImp(InputDialog)

InputDialog::InputDialog( char *prompt,  char *defval, char *retstr) 
{

  // Create simple input dialog.
  fWidgets = new TList;
  
  const TGWindow *main = gClient->GetRoot();
  fuDialog = new TGTransientFrame(main, main, 10, 10);
  
  // command to be executed by buttons and text entry widget
  char *cmd=new char [128];
  sprintf(cmd, "{long r__ptr=0x%x; ((InputDialog*)r__ptr)->ProcessMessage($MSG,$PARM1,$PARM2);}", this);
  
  // create prompt label and textentry widget
  TGLabel *label = new TGLabel(fuDialog, prompt);
  fWidgets->Add(label);
  
  TGTextBuffer *tbuf = new TGTextBuffer(256);  //will be deleted by TGtextEntry
  tbuf->AddText(0, defval);
  
  fTE = new TGTextEntry(fuDialog, tbuf);
  fTE->Resize(260, fTE->GetDefaultHeight());
  fTE->SetCommand(cmd);
  
  TGLayoutHints *l1 = new TGLayoutHints(kLHintsTop | kLHintsLeft, 5, 5, 5, 0);
  TGLayoutHints *l2 = new TGLayoutHints(kLHintsTop | kLHintsLeft, 5, 5, 5, 5);
  fWidgets->Add(l1);
  fWidgets->Add(l2);
  
  fuDialog->AddFrame(label, l1);
  fuDialog->AddFrame(fTE, l2);
  
  // create frame and layout hints for Ok and Cancel buttons
  TGHorizontalFrame *hf = new TGHorizontalFrame(fuDialog, 60, 20, kFixedWidth);
  TGLayoutHints     *l3 = new TGLayoutHints(kLHintsCenterY | kLHintsExpandX, 
					    5, 5, 0, 0);
  
  // put hf as last in list to be deleted
  fWidgets->Add(l3);
  
  // create OK and Cancel buttons in their own frame (hf)
  UInt_t  nb = 0, width = 0, height = 0;
  TGTextButton *b;
  
  b = new TGTextButton(hf, "&Ok", cmd, 1);
  fWidgets->Add(b);
  b->Associate(fuDialog);
  hf->AddFrame(b, l3);
  height = b->GetDefaultHeight();
  width  = TMath::Max(width, b->GetDefaultWidth()); ++nb;
  
  b = new TGTextButton(hf, "&Cancel", cmd, 2);
  fWidgets->Add(b);
  b->Associate(fuDialog);
  hf->AddFrame(b, l3);
  height = b->GetDefaultHeight();
  width  = TMath::Max(width, b->GetDefaultWidth()); ++nb;

  // place button frame (hf) at the bottom
  TGLayoutHints *l4 = new TGLayoutHints(kLHintsBottom | kLHintsCenterX, 
					0, 0, 5, 5);
  fWidgets->Add(l4);
  fWidgets->Add(hf);
  
  fuDialog->AddFrame(hf, l4);
  
  // keep buttons centered and with the same width
  hf->Resize((width + 20) * nb, height);
  
  // set dialog title
  fuDialog->SetWindowName("Get Input");
  
  // map all widgets and calculate size of dialog
  fuDialog->MapSubwindows();
  
  width  = fuDialog->GetDefaultWidth();
  height = fuDialog->GetDefaultHeight();
  
  fuDialog->Resize(width, height);
  
  // position relative to the parent window (which is the root window)
  Window_t wdum;
  int      ax, ay;
  
  gVirtualX->TranslateCoordinates(main->GetId(), main->GetId(),
				  (((TGFrame *) main)->GetWidth()-width) >> 1,
				  (((TGFrame *) main)->GetHeight()-height)>> 1,
				  ax, ay, wdum);
  fuDialog->Move(ax, ay);
  fuDialog->SetWMPosition(ax, ay);
  
  // make the message box non-resizable
  fuDialog->SetWMSize(width, height);
  fuDialog->SetWMSizeHints(width, height, width, height, 0, 0);
  
  fuDialog->SetMWMHints(kMWMDecorAll | kMWMDecorResizeH  | kMWMDecorMaximize |
		       kMWMDecorMinimize | kMWMDecorMenu,
		       kMWMFuncAll  | kMWMFuncResize    | kMWMFuncMaximize |
		       kMWMFuncMinimize,
		       kMWMInputModeless);
  
  // popup dialog and wait till user replies
  fuDialog->MapWindow();
  
  fRetStr = retstr;
  
  gClient->WaitFor(fuDialog);
}

InputDialog::~InputDialog()
{
  // Cleanup dialog.
  
  fWidgets->Delete();
  delete fWidgets;
  
  delete fTE;
  delete fuDialog;
}


Bool_t InputDialog::ProcessMessage(Long_t msg, Long_t parm1, Long_t parm2)
{
  // Handle button and text enter events
  
  printf("\n Handle button and text enter events ") ;getchar() ;
  switch (GET_MSG(msg)) {
  case kC_COMMAND:
    switch (GET_SUBMSG(msg)) {
    case kCM_BUTTON:
      switch (parm1) {
      case 1:
	// here copy the string from text buffer to return variable
	strcpy(fRetStr, fTE->GetBuffer()->GetString());
	delete this;  //CloseWindow() ;
	break;
	
      case 2:
	fRetStr[0] = 0;
	delete this;
	break;
      }
    default:
      break;
    }
    break;
    
  case kC_TEXTENTRY:
    switch (GET_SUBMSG(msg)) {
    case kTE_ENTER:
      // here copy the string from text buffer to return variable
      strcpy(fRetStr, fTE->GetBuffer()->GetString());
      delete this;
      break;
    default:
      break;
    }
    break;
    
  default:
    break;
  }
  return kTRUE;
}



//--- Utility Functions ------------------------------------------------------

char *GetStringDialog( char *prompt,  char *defval)
{
  // Prompt for string. The typed in string is returned.
  
  static char answer[128];
  
  new InputDialog(prompt, defval, answer);
  
  return answer;
}

Int_t GetIntegerDialog( char *prompt, Int_t defval)
{
  // Prompt for integer. The typed in integer is returned.
  
  static char answer[32];
  
  char defv[32];
  sprintf(defv, "%d", defval);
  
  new InputDialog(prompt, defv, answer);
  
  return atoi(answer);
}

Float_t GetFloatDialog( char *prompt, Float_t defval)
{
  // Prompt for float. The typed in float is returned.
  
  static char answer[32];
  
  char defv[32];
  sprintf(defv, "%f", defval);
  
  new InputDialog(prompt, defv, answer);
  
  return atof(answer);
}


char *OpenFileDialog()
{
  // Prompt for file to be opened. Depending on navigation in
  // dialog the current working directory can be changed.
  // The returned file name is always with respect to the
  // current directory.
  
  char *gOpenAsTypes[] = {
    "Macro files",  "*.C",
    "ROOT files",   "*.root",
    "PostScript",   "*.ps",
    "Encapsulated PostScript", "*.eps",
    "Gif files",    "*.gif",
    "All files",    "*",
    0,              0
  };
  
  static TGFileInfo fi;
  fi.fFileTypes = gOpenAsTypes;
  new TGFileDialog(gClient->GetRoot(), gClient->GetRoot(), kFDOpen, &fi);
  
  return fi.fFilename;
}

char *SaveFileDialog()
{
  // Prompt for file to be saved. Depending on navigation in
  // dialog the current working directory can be changed.
  // The returned file name is always with respect to the
  // current directory.
  
  char *gSaveAsTypes[] = {
    "Macro files",  "*.C",
    "ROOT files",   "*.root",
    "PostScript",   "*.ps",
    "Encapsulated PostScript", "*.eps",
    "Gif files",    "*.gif",
    "All files",    "*",
    0,              0
  };
  
  static TGFileInfo fi;
  fi.fFileTypes = gSaveAsTypes;
  new TGFileDialog(gClient->GetRoot(), gClient->GetRoot(), kFDSave, &fi);
  
  return fi.fFilename;
}

